from __future__ import unicode_literals

from cytoolz.functoolz import (
    compose,
    curry,
)

from eth_utils import (
    int_to_big_endian,
)


@curry
def zpad(value, length):
    return value.rjust(length, b'\x00')


zpad32 = zpad(length=32)


int_to_32byte_big_endian = compose(
    zpad32,
    int_to_big_endian,
)



from cytoolz import (
    identity,
)

from eth_utils import (
    decode_hex,
    encode_hex,
    to_canonical_address,
)

from .base import (
    BaseNormalizer,
)
from .inbound import (
    normalize_filter_params as normalize_inbound_filter_params,
    normalize_log_entry as normalize_inbound_log_entry,
    normalize_private_key as normalize_inbound_private_key,
    normalize_raw_transaction as normalize_inbound_raw_transaction,
    normalize_transaction as normalize_inbound_transaction,
)
from .outbound import (
    normalize_account as normalize_outbound_account,
    normalize_account_list as normalize_outbound_account_list,
    normalize_block as normalize_outbound_block,
    normalize_log_entry as normalize_outbound_log_entry,
    normalize_receipt as normalize_outbound_receipt,
    normalize_transaction as normalize_outbound_transaction,
)


class DefaultNormalizer(BaseNormalizer):
    #
    # Inbound
    #
    normalize_inbound_account = staticmethod(to_canonical_address)
    normalize_inbound_block_hash = staticmethod(decode_hex)
    normalize_inbound_block_number = staticmethod(identity)
    normalize_inbound_filter_id = staticmethod(identity)
    normalize_inbound_filter_params = staticmethod(normalize_inbound_filter_params)
    normalize_inbound_log_entry = staticmethod(normalize_inbound_log_entry)
    normalize_inbound_private_key = staticmethod(normalize_inbound_private_key)
    normalize_inbound_raw_transaction = staticmethod(normalize_inbound_raw_transaction)
    normalize_inbound_timestamp = staticmethod(identity)
    normalize_inbound_transaction = staticmethod(normalize_inbound_transaction)
    normalize_inbound_transaction_hash = staticmethod(decode_hex)

    # Outbound
    normalize_outbound_account = staticmethod(normalize_outbound_account)
    normalize_outbound_account_list = staticmethod(normalize_outbound_account_list)
    normalize_outbound_balance = staticmethod(identity)
    normalize_outbound_block_hash = staticmethod(encode_hex)
    normalize_outbound_block = staticmethod(normalize_outbound_block)
    normalize_outbound_code = staticmethod(encode_hex)
    normalize_outbound_filter_id = staticmethod(identity)
    normalize_outbound_log_entry = staticmethod(normalize_outbound_log_entry)
    normalize_outbound_gas_estimate = staticmethod(identity)
    normalize_outbound_nonce = staticmethod(identity)
    normalize_outbound_receipt = staticmethod(normalize_outbound_receipt)
    normalize_outbound_return_data = staticmethod(encode_hex)
    normalize_outbound_transaction = staticmethod(normalize_outbound_transaction)
    normalize_outbound_transaction_hash = staticmethod(encode_hex)



from __future__ import absolute_import

import pkg_resources

from semantic_version import (
    Spec,
)

import rlp

from eth_utils import (
    is_integer,
    to_dict,
    to_tuple,
    to_wei,
)

from eth_tester.constants import (
    FORK_HOMESTEAD,
    FORK_DAO,
    FORK_ANTI_DOS,
    FORK_STATE_CLEANUP,
)
from eth_tester.exceptions import (
    BlockNotFound,
    TransactionNotFound,
    UnknownFork,
    TransactionFailed,
)
from eth_tester.backends.base import BaseChainBackend
from eth_tester.backends.pyethereum.utils import (
    get_pyethereum_version,
    is_pyethereum21_available,
)

from eth_tester.utils.accounts import (
    private_key_to_address,
)

from eth_tester.backends.pyethereum.serializers import (
    serialize_block,
    serialize_transaction,
    serialize_transaction_hash,
    serialize_transaction_receipt,
)
from eth_tester.backends.pyethereum.validation import (
    validate_transaction,
)
from eth_tester.utils.formatting import (
    replace_exceptions,
)


if is_pyethereum21_available():
    from ethereum.tools.tester import (
        TransactionFailed as Pyeth21TransactionFailed,
    )
else:
    Pyeth21TransactionFailed = None


def _get_block_by_number(evm, block_number):
    if is_integer(block_number):
        if block_number > evm.block.number:
            raise BlockNotFound("Block number is longer than current chain.")

        block = evm.chain.get_block_by_number(block_number)
        if block is None:
            raise BlockNotFound("Block not found for block number: {0}".format(block_number))
    elif block_number == "pending":
        block = evm.block  # Get pending block
    elif block_number == "latest":
        block = evm.chain.head  # Get latest block added to chain
    elif block_number == "earliest":
        block = evm.chain.genesis
    else:
        raise BlockNotFound(
            "Block identifier was not in a recognized format: Got "
            "{0}".format(block_number)
        )

    return block


def _get_block_by_hash(evm, block_hash):
    block_by_hash = evm.chain.get_block(block_hash)
    if block_by_hash is None or block_by_hash.number == evm.block.number:
        raise BlockNotFound(
            "Block with hash {0} not found on chain".format(block_hash)
        )

    block_by_number = _get_block_by_number(evm, block_by_hash.number)
    if block_by_hash.hash != block_by_number.hash:
        raise BlockNotFound(
            "Block with hash {0} not found on chain".format(block_hash)
        )
    return block_by_hash


EMPTY_RECEIPTS_ROOT = b'V\xe8\x1f\x17\x1b\xccU\xa6\xff\x83E\xe6\x92\xc0\xf8n\x5bH\xe0\x1b\x99l\xad\xc0\x01b/\xb5\xe3c\xb4!'  # noqa: E501


def _get_state_by_block_hash(evm, block_hash, ephemeral=False):
    from ethereum.messages import Receipt
    if block_hash == evm.block.hash:
        block = evm.block
        base_state = evm.head_state
    else:
        block = _get_block_by_hash(evm, block_hash)
        base_state = evm.chain.mk_poststate_of_blockhash(block_hash)

    if ephemeral:
        state = base_state.ephemeral_clone()
    else:
        state = base_state

    if block.header.receipts_root != EMPTY_RECEIPTS_ROOT:
        receipt_list = rlp.decode(evm.chain.db.get(block.header.receipts_root))
        state.receipts = [
            rlp.decode(rlp.encode(receipt_items), Receipt)
            for receipt_items
            in receipt_list
        ]
    return state


def _get_state_by_block_number(evm, block_number, ephemeral=False):
    block = _get_block_by_number(evm, block_number)
    return _get_state_by_block_hash(evm, block.hash, ephemeral)


def _get_transaction_by_hash(evm, transaction_hash):
    for index, candidate in enumerate(evm.block.transactions):
        if transaction_hash == candidate.hash:
            return (
                evm.block,
                candidate,
                index,
                True,
            )

    for block_number in range(evm.chain.head.number, -1, -1):
        block = _get_block_by_number(evm, block_number)
        for index, transaction in enumerate(block.transactions):
            if transaction.hash == transaction_hash:
                return block, transaction, index, False
    else:
        raise TransactionNotFound(
            "Transaction with hash {0} not found".format(
                transaction_hash,
            )
        )


def _get_key_for_account(evm, account):
    from ethereum.tools import tester
    if account in tester.accounts:
        index = tester.accounts.index(account)
        return tester.keys[index]
    elif account in evm.extra_accounts:
        return evm.extra_accounts[account]
    else:
        raise KeyError("Account {:#x} not found in known accounts")


@to_dict
def _format_transaction(evm, transaction):
    if 'from' in transaction:
        yield 'sender', _get_key_for_account(evm, transaction['from'])
    if 'gas' in transaction:
        yield 'startgas', transaction['gas']
    if 'gas_price' in transaction:
        yield 'gasprice', transaction['gas_price']

    if 'to' in transaction:
        yield 'to', transaction['to']
    else:
        yield 'to', b''

    if 'value' in transaction:
        yield 'value', transaction['value']
    if 'data' in transaction:
        yield 'data', transaction['data']


def _send_transaction(evm, raw_transaction):
    validate_transaction(raw_transaction)

    transaction = _format_transaction(evm, raw_transaction)
    output = evm.tx(**transaction)
    return output


class PyEthereum21Backend(BaseChainBackend):
    def __init__(self):
        if not is_pyethereum21_available():
            version = get_pyethereum_version()
            if version is None:
                raise pkg_resources.DistributionNotFound(
                    "The `ethereum` package is not available.  The "
                    "`PyEthereum21Backend` requires a 2.0.0+ version of the "
                    "ethereum package to be installed."
                )
            elif version not in Spec('>=2.0.0,<2.2.0'):
                raise pkg_resources.DistributionNotFound(
                    "The `PyEthereum21Backend` requires a 2.0.0+ version of the "
                    "`ethereum` package.  Found {0}".format(version)
                )
        self.reset_to_genesis()

    #
    # Fork block numbers
    #
    def set_fork_block(self, fork_name, fork_block):
        if fork_name == FORK_HOMESTEAD:
            self.evm.chain.env.config['HOMESTEAD_FORK_BLKNUM'] = fork_block
        elif fork_name == FORK_DAO:
            # NOTE: REALLY WEIRD HACK to get the dao_fork_blk to accept block 0
            if not fork_block:
                self.evm.chain.env.config['DAO_FORK_BLKNUM'] = 999999999999999
            else:
                self.evm.chain.env.config['DAO_FORK_BLKNUM'] = fork_block
        elif fork_name == FORK_ANTI_DOS:
            self.evm.chain.env.config['ANTI_DOS_FORK_BLKNUM'] = fork_block
        elif fork_name == FORK_STATE_CLEANUP:
            self.evm.chain.env.config['CLEARING_FORK_BLKNUM'] = fork_block
        else:
            raise UnknownFork("Unknown fork name: {0}".format(fork_name))

    def get_fork_block(self, fork_name):
        if fork_name == FORK_HOMESTEAD:
            return self.evm.chain.env.config['HOMESTEAD_FORK_BLKNUM']
        elif fork_name == FORK_DAO:
            return self.evm.chain.env.config['DAO_FORK_BLKNUM']
        elif fork_name == FORK_ANTI_DOS:
            return self.evm.chain.env.config['ANTI_DOS_FORK_BLKNUM']
        elif fork_name == FORK_STATE_CLEANUP:
            return self.evm.chain.env.config['CLEARING_FORK_BLKNUM']
        else:
            raise UnknownFork("Unknown fork name: {0}".format(fork_name))

    #
    # Snapshot API
    #
    def take_snapshot(self):
        self.evm.head_state.commit()
        block = _get_block_by_number(self.evm, 'latest')
        return block.hash

    def revert_to_snapshot(self, snapshot):
        self.evm.change_head(snapshot)

        latest_state = _get_state_by_block_hash(self.evm, snapshot)
        if latest_state.block_number > 0:
            latest_state.block_number += 1

        self.evm.chain.state = latest_state
        self.evm.chain.head_hash = snapshot

    def reset_to_genesis(self):
        from ethereum.tools import tester
        self.evm = tester.Chain()
        self.evm = tester.Chain(alloc={
            account: {'balance': to_wei(1000000, 'ether')}
            for account
            in tester.accounts
        })
        self.evm.extra_accounts = {}

    #
    # Meta
    #
    def time_travel(self, to_timestamp):
        assert self.evm.block.header.timestamp < to_timestamp
        self.evm.block.header.timestamp = to_timestamp - 1
        self.mine_blocks()

    #
    # Mining
    #
    @to_tuple
    def mine_blocks(self, num_blocks=1, coinbase=None):
        if not coinbase:
            coinbase = self.get_accounts()[0]

        from ethereum.common import mk_receipt_sha

        for _ in range(num_blocks):
            receipts = self.evm.head_state.receipts
            block = self.evm.mine(coinbase=coinbase)
            if block is None:
                # earlier versions of pyethereum21 didn't return the block.
                block = _get_block_by_number(self.evm, self.evm.block.number - 1)

            receipts_root = mk_receipt_sha(receipts)
            self.evm.chain.db.put(receipts_root, rlp.encode(receipts))

            yield block.hash

    #
    # Accounts
    #
    def get_key_for_account(self, account):
        from ethereum.tools import tester
        if account in tester.accounts:
            index = tester.accounts.index(account)
            return tester.keys[index]
        elif account in self.evm.extra_accounts:
            return self.evm.extra_accounts[account]
        else:
            raise KeyError("Account {:#x} not found in known accounts")

    @to_tuple
    def get_accounts(self):
        from ethereum.tools import tester
        for account in tester.accounts:
            yield account

        for account in self.evm.extra_accounts.keys():
            yield account

    def add_account(self, private_key):
        account = private_key_to_address(private_key)
        self.evm.extra_accounts[account] = private_key

    #
    # Chain data
    #
    def get_block_by_number(self, block_number, full_transactions=False):
        block = _get_block_by_number(self.evm, block_number)

        if full_transactions:
            transaction_serialize_fn = serialize_transaction
        else:
            transaction_serialize_fn = serialize_transaction_hash

        is_pending = block == self.evm.block
        return serialize_block(self.evm, block, transaction_serialize_fn, is_pending)

    def get_block_by_hash(self, block_hash, full_transactions=False):
        if full_transactions:
            transaction_serialize_fn = serialize_transaction
        else:
            transaction_serialize_fn = serialize_transaction_hash

        block = _get_block_by_hash(self.evm, block_hash)
        assert block is not None, "Block not found! Given {:#x}".format(block_hash)

        is_pending = block == self.evm.block
        return serialize_block(self.evm, block, transaction_serialize_fn, is_pending)

    def get_transaction_by_hash(self, transaction_hash):
        block, transaction, transaction_index, is_pending = _get_transaction_by_hash(
            self.evm,
            transaction_hash,
        )
        return serialize_transaction(block, transaction, transaction_index, is_pending)

    def get_transaction_receipt(self, transaction_hash):
        block, transaction, transaction_index, is_pending = _get_transaction_by_hash(
            self.evm,
            transaction_hash,
        )
        state = _get_state_by_block_hash(self.evm, block.hash)
        return serialize_transaction_receipt(
            block,
            transaction,
            state.receipts[transaction_index],
            transaction_index,
            is_pending,
        )

    #
    # Account state
    #
    # NOTE: Added as a helper, might be more broadly useful
    def get_nonce(self, account, block_number="latest"):
        state = _get_state_by_block_number(self.evm, block_number)
        return state.get_nonce(account)

    def get_balance(self, account, block_number="latest"):
        state = _get_state_by_block_number(self.evm, block_number)
        return state.get_balance(account)

    def get_code(self, account, block_number="latest"):
        state = _get_state_by_block_number(self.evm, block_number)
        return state.get_code(account)

    #
    # Transactions
    #
    def send_raw_transaction(self, raw_transaction):
        from ethereum.transactions import Transaction
        rlp_transaction = rlp.decode(raw_transaction, Transaction)
        self.evm.direct_tx(rlp_transaction)
        return self.evm.last_tx.hash

    def send_transaction(self, transaction):
        _send_transaction(self.evm, transaction)
        return self.evm.last_tx.hash

    @replace_exceptions({Pyeth21TransactionFailed: TransactionFailed})
    def estimate_gas(self, transaction):
        snapshot = self.take_snapshot()
        _send_transaction(self.evm, transaction)
        gas_used = self.evm.head_state.receipts[-1].gas_used
        self.revert_to_snapshot(snapshot)
        return gas_used

    @replace_exceptions({Pyeth21TransactionFailed: TransactionFailed})
    def call(self, transaction, block_number="latest"):
        from ethereum.messages import apply_message

        if isinstance(block_number, bytes):
            state = self.evm.chain.mk_poststate_of_blockhash(block_number)
        elif isinstance(block_number, int) or isinstance(block_number, str):
            block = _get_block_by_number(self.evm, block_number)
            state = self.evm.chain.mk_poststate_of_blockhash(block.hash)
        else:
            raise BlockNotFound("Invalid block identifer.")

        output = apply_message(
            state,
            sender=transaction['from'],
            to=transaction['to'],
            code_address=transaction['to'],
            data=transaction['data'],
            gas=transaction['gas'],
        )
        if output is None:
            raise TransactionFailed
        return output



from eth_utils import (
    to_tuple,
    is_bytes,
    is_address,
    is_same_address,
    is_integer,
)

from queue import (
    Queue,
    Empty,
)


class Filter(object):
    filter_params = None
    filter_fn = None

    values = None
    queue = None

    def __init__(self, filter_params, filter_fn=None):
        self.filter_params = filter_params
        self.filter_fn = filter_fn

        self.values = []
        self.queue = Queue()

    @to_tuple
    def get_changes(self):
        while True:
            try:
                yield self.queue.get_nowait()
            except Empty:
                break

    def get_all(self):
        return tuple(self.values)

    def add(self, *values):
        for item in values:
            if self.filter_fn is not None and not self.filter_fn(item):
                continue
            self.values.append(item)
            self.queue.put_nowait(item)

    def remove(self, *values):
        try:
            values_to_remove = set(values)
        except TypeError:
            # log filters are dicts which are not hashable
            values_to_remove = values

        queued_values = self.get_changes()
        self.values = [
            value
            for value
            in self.get_all()
            if value not in values_to_remove
        ]
        for value in queued_values:
            if value in values_to_remove:
                continue
            self.queue.put_nowait(value)


def is_tuple(value):
    return isinstance(value, tuple)


def is_topic_string(value):
    return is_bytes(value) and len(value) == 32


def is_topic(value):
    return value is None or is_topic_string(value)


def is_flat_topic_array(value):
    return is_tuple(value) and all(is_topic(item) for item in value)


def is_nested_topic_array(value):
    return bool(value) and is_tuple(value) and all((is_topic_array(item) for item in value))


def is_topic_array(value):
    return is_flat_topic_array(value) or is_nested_topic_array(value)


def check_single_topic_match(log_topic, filter_topic):
    if filter_topic is None:
        return True
    # python2 thinks string and bytes values can be equal.
    return filter_topic == log_topic and type(log_topic) is type(filter_topic)


def check_if_from_block_match(block_number, _type, from_block):
    if from_block is None or from_block == "latest":
        return _type == "mined"
    elif from_block in {"earliest", "pending"}:
        return _type == "pending"
    elif is_integer(from_block):
        return is_integer(block_number) and block_number >= from_block
    else:
        raise ValueError("Unrecognized from_block format: {0}".format(from_block))


def check_if_to_block_match(block_number, _type, to_block):
    if to_block is None or to_block == "latest":
        return _type == "mined"
    elif to_block in {"earliest", "pending"}:
        return _type == "pending"
    elif is_integer(to_block):
        return is_integer(block_number) and block_number <= to_block
    else:
        raise ValueError("Unrecognized to_block format: {0}".format(to_block))


def check_if_log_matches_flat_topics(log_topics, filter_topics):
    if not filter_topics:
        return True
    elif len(log_topics) < len(filter_topics):
        return False
    else:
        return all(
            check_single_topic_match(left, right)
            for left, right
            in zip(log_topics, filter_topics)
        )


def check_if_topics_match(log_topics, filter_topics):
    if filter_topics is None:
        return True
    elif is_flat_topic_array(filter_topics):
        return check_if_log_matches_flat_topics(log_topics, filter_topics)
    elif is_nested_topic_array(filter_topics):
        return any(
            check_if_log_matches_flat_topics(log_topics, sub_filter_topics)
            for sub_filter_topics
            in filter_topics
        )
    else:
        raise ValueError("Unrecognized topics format: {0}".format(filter_topics))


def check_if_address_match(address, addresses):
    if addresses is None:
        return True
    if is_tuple(addresses):
        return any(
            is_same_address(address, item)
            for item
            in addresses
        )
    elif is_address(addresses):
        return is_same_address(addresses, address)
    else:
        raise ValueError("Unrecognized address format: {0}".format(addresses))


def check_if_log_matches(log_entry,
                         from_block,
                         to_block,
                         addresses,
                         topics):
    if not check_if_from_block_match(log_entry['block_number'], log_entry['type'], from_block):
        return False
    elif not check_if_to_block_match(log_entry['block_number'], log_entry['type'], to_block):
        return False
    elif not check_if_address_match(log_entry['address'], addresses):
        return False
    elif not check_if_topics_match(log_entry['topics'], topics):
        return False
    else:
        return True



from __future__ import absolute_import

import pkg_resources

from semantic_version import (
    Version,
    Spec,
)


def get_pyethereum_version():
    try:
        return Version(pkg_resources.get_distribution("ethereum").version)
    except pkg_resources.DistributionNotFound:
        return None


def is_pyethereum16_available():
    pyethereum_version = get_pyethereum_version()

    if pyethereum_version is None:
        return False
    elif pyethereum_version not in Spec('>=1.6.0,<1.7.0'):
        return False
    else:
        return True


def is_pyethereum21_available():
    pyethereum_version = get_pyethereum_version()

    if pyethereum_version is None:
        return False
    elif pyethereum_version not in Spec('>=2.1.0,<2.2.0'):
        return False
    else:
        return True



from __future__ import absolute_import

from cytoolz.functoolz import (
    partial,
    identity,
    compose,
)

from eth_utils.curried import (
    apply_one_of_formatters,
    to_checksum_address,
    encode_hex,
    is_address,
    is_bytes,
    is_canonical_address,
    is_dict,
)

from .common import (
    normalize_if,
    normalize_dict,
    normalize_array,
)


normalize_account = to_checksum_address
normalize_account_list = partial(normalize_array, normalizer=normalize_account)

to_empty_or_checksum_address = apply_one_of_formatters((
    (lambda addr: addr == b'', lambda addr: ''),
    (is_canonical_address, to_checksum_address),
))

TRANSACTION_NORMALIZERS = {
    "hash": encode_hex,
    "nonce": identity,
    "block_hash": partial(normalize_if, conditional_fn=is_bytes, normalizer=encode_hex),
    "block_number": identity,
    "transaction_index": identity,
    "from": to_checksum_address,
    "to": to_empty_or_checksum_address,
    "value": identity,
    "gas": identity,
    "gas_price": identity,
    "data": encode_hex,
    "v": identity,
    "r": identity,
    "s": identity,
}


normalize_transaction = partial(normalize_dict, normalizers=TRANSACTION_NORMALIZERS)


def is_transaction_hash_list(value):
    return all(is_bytes(item) for item in value)


def is_transaction_object_list(value):
    return all(is_dict(item) for item in value)


BLOCK_NORMALIZERS = {
    "number": identity,
    "hash": encode_hex,
    "parent_hash": encode_hex,
    "nonce": encode_hex,
    "sha3_uncles": encode_hex,
    "logs_bloom": identity,
    "transactions_root": encode_hex,
    "receipts_root": encode_hex,
    "state_root": encode_hex,
    "miner": to_checksum_address,
    "difficulty": identity,
    "total_difficulty": identity,
    "size": identity,
    "extra_data": encode_hex,
    "gas_limit": identity,
    "gas_used": identity,
    "timestamp": identity,
    "transactions": compose(
        partial(
            normalize_if,
            conditional_fn=is_transaction_hash_list,
            normalizer=partial(normalize_array, normalizer=encode_hex),
        ),
        partial(
            normalize_if,
            conditional_fn=is_transaction_object_list,
            normalizer=partial(normalize_array, normalizer=normalize_transaction),
        ),
    ),
    "uncles": partial(normalize_array, normalizer=encode_hex),
}


normalize_block = partial(normalize_dict, normalizers=BLOCK_NORMALIZERS)


LOG_ENTRY_NORMALIZERS = {
    "type": identity,
    "log_index": identity,
    "transaction_index": partial(
        normalize_if,
        conditional_fn=is_bytes,
        normalizer=encode_hex,
    ),
    "transaction_hash": encode_hex,
    "block_hash": partial(
        normalize_if,
        conditional_fn=is_bytes,
        normalizer=encode_hex,
    ),
    "block_number": identity,
    "address": to_checksum_address,
    "data": encode_hex,
    "topics": partial(normalize_array, normalizer=encode_hex),
}


normalize_log_entry = partial(normalize_dict, normalizers=LOG_ENTRY_NORMALIZERS)


RECEIPT_NORMALIZERS = {
    "transaction_hash": encode_hex,
    "transaction_index": identity,
    "block_number": identity,
    "block_hash": partial(
        normalize_if,
        conditional_fn=is_bytes,
        normalizer=encode_hex,
    ),
    "cumulative_gas_used": identity,
    "gas_used": identity,
    "contract_address": partial(
        normalize_if,
        conditional_fn=is_address,
        normalizer=to_checksum_address,
    ),
    "logs": partial(normalize_array, normalizer=normalize_log_entry),
}


normalize_receipt = partial(normalize_dict, normalizers=RECEIPT_NORMALIZERS)



class BaseChainBackend(object):
    #
    # Snapshot API
    #
    def take_snapshot(self):
        raise NotImplementedError("Must be implemented by subclasses")

    def revert_to_snapshot(self, snapshot):
        raise NotImplementedError("Must be implemented by subclasses")

    def reset_to_genesis(self):
        raise NotImplementedError("Must be implemented by subclasses")

    #
    # Fork block numbers
    #
    def set_fork_block(self, fork_name, fork_block):
        raise NotImplementedError("Must be implemented by subclasses")

    def get_fork_block(self, fork_name):
        raise NotImplementedError("Must be implemented by subclasses")

    #
    # Meta
    #
    def time_travel(self, to_timestamp):
        raise NotImplementedError("Must be implemented by subclasses")

    #
    # Mining
    #
    def mine_blocks(self, num_blocks=1, coinbase=None):
        raise NotImplementedError("Must be implemented by subclasses")

    #
    # Accounts
    #
    def get_accounts(self):
        raise NotImplementedError("Must be implemented by subclasses")

    def add_account(self, private_key):
        raise NotImplementedError("Must be implemented by subclasses")

    #
    # Chain data
    #
    def get_block_by_number(self, block_number, full_transaction=True):
        raise NotImplementedError("Must be implemented by subclasses")

    def get_block_by_hash(self, block_hash, full_transaction=True):
        raise NotImplementedError("Must be implemented by subclasses")

    def get_transaction_by_hash(self, transaction_hash):
        raise NotImplementedError("Must be implemented by subclasses")

    def get_transaction_receipt(self, transaction_hash):
        raise NotImplementedError("Must be implemented by subclasses")

    #
    # Account state
    #
    def get_nonce(self, account, block_number=None):
        raise NotImplementedError("Must be implemented by subclasses")

    def get_balance(self, account, block_number=None):
        raise NotImplementedError("Must be implemented by subclasses")

    def get_code(self, account, block_number=None):
        raise NotImplementedError("Must be implemented by subclasses")

    #
    # Transactions
    #
    def send_transaction(self, transaction):
        raise NotImplementedError("Must be implemented by subclasses")

    def send_signed_transaction(self, transaction):
        raise NotImplementedError("Must be implemented by subclasses")

    def estimate_gas(self, transaction):
        raise NotImplementedError("Must be implemented by subclasses")

    def call(self, transaction, block_number="latest"):
        raise NotImplementedError("Must be implemented by subclasses")



from __future__ import unicode_literals

import binascii

from cytoolz import (
    partial,
)

from eth_utils import (
    is_boolean,
    is_checksum_address,
    is_checksum_formatted_address,
    is_dict,
    is_hex,
    is_hex_address,
    is_integer,
    is_list_like,
    is_string,
    is_text,
    remove_0x_prefix,
    decode_hex,
)

from eth_tester.constants import (
    BLOCK_NUMBER_META_VALUES,
)
from eth_tester.exceptions import (
    ValidationError,
)

from .common import (
    validate_positive_integer,
    validate_uint256,
    validate_uint8,
    validate_text,
)


def is_32byte_hex_string(value):
    return is_text(value) and is_hex(value) and len(remove_0x_prefix(value)) == 64


def is_topic(value):
    return value is None or is_32byte_hex_string(value)


def validate_32_byte_hex_value(value, name):
    error_message = (
        "{0} must be a hexidecimal encoded 32 byte string.  Got: "
        "{1}".format(name, value)
    )
    if not is_32byte_hex_string(value):
        raise ValidationError(error_message)


MAX_TIMESTAMP = 33040162800  # January 1st 3017 is appropriately far in the future.


def validate_timestamp(value):
    validate_positive_integer(value)

    if value >= MAX_TIMESTAMP:
        raise ValidationError(
            "Timestamp values must be less than {0}.  Got {1}".format(
                MAX_TIMESTAMP,
                value,
            )
        )


def validate_block_number(value):
    error_message = (
        "Block number must be a positive integer or one of the strings "
        "'latest', 'earliest', or 'pending'.  Got: {0}".format(value)
    )
    if is_string(value):
        validate_text(value)
        if value not in BLOCK_NUMBER_META_VALUES:
            raise ValidationError(error_message)
    elif not is_integer(value) or is_boolean(value):
        raise ValidationError(error_message)
    elif value < 0:
        raise ValidationError(error_message)


validate_block_hash = partial(validate_32_byte_hex_value, name="Block hash")
validate_transaction_hash = partial(validate_32_byte_hex_value, name="Transaction hash")
validate_filter_id = partial(validate_positive_integer)


def validate_account(value):
    if not is_text(value):
        raise ValidationError("Address must be 20 bytes encoded as hexidecimal")
    elif not is_hex_address(value):
        raise ValidationError("Address must be 20 bytes encoded as hexidecimal")
    elif is_checksum_formatted_address(value) and not is_checksum_address(value):
        raise ValidationError("Address does not validate EIP55 checksum")


def is_flat_topic_array(value):
    if not is_list_like(value):
        return False
    return all(is_topic(item) for item in value)


def validate_filter_params(from_block, to_block, address, topics):
    # blocks
    if from_block is not None:
        validate_block_number(from_block)
    if to_block is not None:
        validate_block_number(to_block)

    # address
    if address is None:
        pass
    elif is_list_like(address):
        if not address:
            raise ValidationError(
                "Address must be either a single hexidecimal encoded address or "
                "a non-empty list of hexidecimal encoded addresses"
            )
        for sub_address in address:
            validate_account(sub_address)
    elif not is_hex_address(address):
        validate_account(address)

    invalid_topics_message = (
        "Topics must be one of `None`, an array of 32 byte hexidecimal encoded "
        "strings, or an array of arrays of 32 byte hexidecimal strings"
    )
    # topics
    if topics is None:
        pass
    elif not is_list_like(topics):
        raise ValidationError(invalid_topics_message)
    elif is_flat_topic_array(topics):
        return True
    elif all(is_flat_topic_array(item) for item in topics):
        return True
    else:
        raise ValidationError(invalid_topics_message)


def validate_private_key(value):
    if not is_text(value) or not is_hex(value) or not len(remove_0x_prefix(value)) == 64:
        raise ValidationError("Private keys must be 32 bytes encoded as hexidecimal")


TRANSACTION_KEYS = {
    'from',
    'to',
    'gas',
    'gas_price',
    'value',
    'data',
    'nonce',
}

SIGNED_TRANSACTION_KEYS = {
    'r',
    's',
    'v',
}

TRANSACTION_TYPE_INFO = {
    'send': TRANSACTION_KEYS,
    'send_signed': TRANSACTION_KEYS.union(SIGNED_TRANSACTION_KEYS),
    'call': TRANSACTION_KEYS.difference({'nonce'}),
    'estimate': TRANSACTION_KEYS.difference({'nonce'}),
}

ALLOWED_TRANSACTION_TYPES = set(TRANSACTION_TYPE_INFO.keys())


def validate_transaction(value, txn_type):
    if txn_type not in ALLOWED_TRANSACTION_TYPES:
        raise TypeError("the `txn_type` parameter must be one of send/call/estimate")
    if not is_dict(value):
        raise ValidationError("Transaction must be a dictionary.  Got: {0}".format(type(value)))

    unknown_keys = tuple(sorted(set(value.keys()).difference(
        TRANSACTION_TYPE_INFO[txn_type],
    )))
    if unknown_keys:
        raise ValidationError(
            "Only the keys '{0}' are allowed.  Got extra keys: '{1}'".format(
                "/".join(tuple(sorted(TRANSACTION_TYPE_INFO[txn_type]))),
                "/".join(unknown_keys),
            )
        )

    if txn_type == 'send':
        required_keys = {'from', 'gas'}
    elif txn_type == 'send_signed':
        required_keys = {'from', 'gas'} | SIGNED_TRANSACTION_KEYS
    elif txn_type in {'estimate', 'call'}:
        required_keys = set(['from'])
    else:
        raise Exception("Invariant: code path should be unreachable")

    missing_required_keys = tuple(sorted(required_keys.difference(value.keys())))
    if missing_required_keys:
        raise ValidationError(
            "Transaction is missing the required keys: '{0}'".format(
                "/".join(missing_required_keys),
            )
        )

    if 'from' in value:
        validate_account(value['from'])

    if 'to' in value and value['to'] != '':
        validate_account(value['to'])
    elif 'to' in value and value['to'] == '':
        validate_text(value['to'])

    if 'gas' in value:
        validate_uint256(value['gas'])

    if 'gas_price' in value:
        validate_uint256(value['gas_price'])

    if 'value' in value:
        validate_uint256(value['value'])

    if 'nonce' in value:
        validate_uint256(value['nonce'])

    if 'data' in value:
        bad_data_message = (
            "Transaction data must be a hexidecimal encoded string.  Got: "
            "{0}".format(value['data'])
        )
        if not is_text(value['data']):
            raise ValidationError(bad_data_message)
        elif not remove_0x_prefix(value['data']):
            pass
        elif not is_hex(value['data']):
            raise ValidationError(bad_data_message)
        try:
            decode_hex(value['data'])
        except (binascii.Error, TypeError):
            # TypeError is for python2
            # binascii.Error is for python3
            raise ValidationError(bad_data_message)

    if txn_type == 'send_signed':
        validate_uint256(value['r'])
        validate_uint256(value['s'])
        validate_uint8(value['v'])


def validate_raw_transaction(raw_transaction):
    if not is_text(raw_transaction) or not is_hex(raw_transaction):
        raise ValidationError(
            "Raw Transaction must be a hexidecimal encoded string.  Got: "
            "{0}".format(raw_transaction)
        )



from cytoolz.dicttoolz import (
    assoc,
)
from cytoolz.functoolz import (
    pipe,
    partial,
)


def serialize_block(block, transaction_serializer, is_pending):
    serialized_transactions = tuple(
        transaction_serializer(transaction, block, transaction_index, is_pending=is_pending)
        for transaction_index, transaction
        in enumerate(block['transactions'])
    )
    return assoc(block, 'transactions', serialized_transactions)


def serialize_transaction_as_hash(transaction, block, transaction_index, is_pending):
    return transaction['hash']


def serialize_full_transaction(transaction, block, transaction_index, is_pending):
    if is_pending:
        block_number = None
        block_hash = None
        transaction_index = None
    else:
        block_number = block['number']
        block_hash = block['hash']

    return pipe(
        transaction,
        partial(assoc, key='block_number', value=block_number),
        partial(assoc, key='block_hash', value=block_hash),
        partial(assoc, key='transaction_index', value=transaction_index),
    )


def serialize_receipt(transaction, block, transaction_index, is_pending):
    if is_pending:
        block_number = None
        block_hash = None
        transaction_index = None
    else:
        block_number = block['number']
        block_hash = block['hash']

    return pipe(
        transaction,
        partial(assoc, key='block_number', value=block_number),
        partial(assoc, key='block_hash', value=block_hash),
        partial(assoc, key='transaction_index', value=transaction_index),
    )



# TODO: Should this be moved to a common package like eth-utils?
from eth_utils import (
    to_list,
)


VALID_TRANSACTION_PARAMS = [
    'from',
    'to',
    'gas',
    'gas_price',
    'value',
    'data',
    'nonce',
    'r',
    's',
    'v',
]


def extract_valid_transaction_params(transaction_params):
    return {key: transaction_params[key]
            for key in VALID_TRANSACTION_PARAMS if key in transaction_params}


@to_list
def remove_matching_transaction_from_list(transaction_list, transaction):
    for tx in transaction_list:
        nonce_equal = transaction['nonce'] == tx['nonce']
        from_equal = transaction['from'] == tx['from']
        match = nonce_equal and from_equal
        if not match:
            yield tx



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import (
    setup,
    find_packages,
)


setup(
    name='eth-tester',
    # *IMPORTANT*: Don't manually change the version here. Use the 'bumpversion' utility.
    version='0.1.0-beta.21',
    description="""Tools for testing Ethereum applications.""",
    long_description_markdown_filename='README.md',
    author='Piper Merriam',
    author_email='pipermerriam@gmail.com',
    url='https://github.com/ethereum/eth-tester',
    include_package_data=True,
    install_requires=[
        "cytoolz>=0.9.0,<1.0.0",
        "eth-utils>=1.0.1,<2.0.0",
        "rlp>=0.6.0,<1.0.0",
        "semantic_version>=2.6.0,<3.0.0",
        "eth-keys>=0.2.0-beta.2,<0.3.0",
    ],
    extras_require={
        'pyethereum16': [
            "ethereum>=1.6.0,<2.0.0",
        ],
        'pyethereum21': [
            "ethereum>=2.1.0,<2.2.0",
        ],
        'py-evm': [
            "py-evm>=0.2.0a11,<1.0.0",  # evm is very high velocity and might change API at each alpha
        ],
    },
    setup_requires=['setuptools-markdown'],
    py_modules=['eth_tester'],
    license="MIT",
    zip_safe=False,
    keywords='ethereum',
    packages=find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)
